#define RS_ASM
        TITLE     "Unit test for RS-232 routines"
		include "pic.inc"
		include "board.inc"
		include "rs.inc"
		
;---------------------------------------------------------------------------
;      Copyright: Eugene M. Hutorny  2003,eugene@ksf.kiev.ua
;      Revision:  V0.01
;      Date:      Feb 25, 2003
;      Assembler: MPASM version 3.20.07
;---------------------------------------------------------------------------
	__CONFIG	_WDT_ON & _XT_OSC
	__IDLOCS	0x01130

; --------------------------------------------------------------------------
; Application specific RS-232 definitions
;---------------------------------------------------------------------------
#define RS_BUFFSIZE .57
#define RS_MAXSKID  .16
#define RS_BITCOUNT .8
#define	RS_INTCON	B'00000101'				; disable all interrupts and clear INTF
; GIE ----------------+|||||||
; EEIE ----------------+||||||
; T0IE -----------------+|||||
; INTE ------------------+||||
; RBIE -------------------+|||
; T0IF --------------------+||
; INTF ---------------------+|
; RBIF ----------------------+

	UDATA
;---------------------------------------------------------------------------
; UI file registers
;---------------------------------------------------------------------------
UI_SELECTION	res 1			; speed selection detected on boot up
UI_TICK			res 1			; tick counter for blinking LEDs
UI_BLINKER		res 1			; blink mask to toggle LEDs

;---------------------------------------------------------------------------
; Application registers
;---------------------------------------------------------------------------
#define	AP_SPEED4		UI_SELECTION, 7
#define	AP_SPEED3		UI_SELECTION, 6
#define	AP_SPEED2		UI_SELECTION, 5
#define	AP_SPEED1		UI_SELECTION, 4
#define	UI_ERROR		UI_SELECTION, 3

;AP_FLAGS	 	res 1			; RS & AP status flags
TMP_WREG		res 1			; temporary storage for WREG
TMP_STATUS		res 1			;          - " -        STATUS
TMP_INTCON		res	1			;		   - " -		INTCON
AP_DATA			res 1			; data for sending out

; --------------------------------------------------------------------------
; RS file registers
;---------------------------------------------------------------------------
RS_DATA		 	res 1			; file register used in RS
RS_COUNT		res 1			; bit counter
RS_ALIGN		res 1			; alignment bits
RS_COUNTER		res 1			; recieved/echoed counter
RS_BUFFER		res RS_BUFFSIZE	; data buffer

RS_setBUFFER macro
	movlw	RS_BUFFER
	movwf	FSR
  endm

;---------------------------------------------------------------------------
; Application code
;---------------------------------------------------------------------------
	CODE 	0x00 			; Reset vector
	btfss	STATUS, NOT_TO	; Timeout detected
	goto    TIMEOUT			; Jump to Timeout routine
	goto	START			; Jump to start
IV	CODE    0x04
	movwf	TMP_WREG				; 4: save W
V	CODE	0x05	
	movlw	0						; 5: safely clear W
	btfsc	RX				; 6: test port RX
	movlw	1				; 7: set bit 0
	INT_skpIF						; 8: now we have some time to look around what caused this interrupt
	goto	not_INT					; 9: not INT
	movwf	INDF					; 10: save what already recived to buffer 
	btfsc	RX				; 11: test port
	bsf		INDF,1			; 12: set bit 1
	btfss	AP_SPEED4				; 13: test if is the selected speed is 225K
	goto	INT_SPEED3				; 14: no, it is not
	btfsc	RX				; 15: test port
	bsf		INDF,2			; 16: set bit 2
	swapf	STATUS, W				; 17: load STATUS
	movwf	TMP_STATUS				; 18: save STATUS
	btfsc	RX				; 19: test port
	bsf		INDF,3			; 20: set bit
	incf	RS_COUNTER, F			; 21: already safe to increment counter
	btfsc	RS_COUNTER, .5			; 22: test if counter exceeds .32
	CTS_setMARK						; 23: request stop transferring	
	btfsc	RX				; 24: test port
	bsf		INDF,4			; 25: set bit
	nope							; 26, 27 still have plenty of time
	btfsc	RX				; 28: test port
	bsf		INDF,5			; 29: set bit
	movf	TMP_WREG, W				; 30: W is not used we may restore WREG
	TMR0_clr						; 31: clear timer, used for timing out between bytes
	btfsc	RX				; 32: test port
	bsf		INDF,6			; 33: set bit	
	nope							; 34, 35
	nop								; 36
	btfsc	RX				; 37: test port
	bsf		INDF,7			; 38: set the last bit
	incf	FSR, F					; 39: advance current position, Z is discarded with 0 (always FSR > 0)
	btfsc	TMP_STATUS, (Z+.4)		; 40: what was the last saved value of Z? (+.4 because of SWAP)
	setz							; 41: restore Z, it was the only flag discarded in the routine
	INT_cf							; 42: interrupt handled
	retfie							; 43: leave routine - 4 cycles of two stop bits left !!!
		
not_INT
	swapf	STATUS, W				; load STATUS
	movwf	TMP_STATUS				; save STATUS
	TMR0_skpc						; this must be TMR0 interupt
	call	UI_BLINK	
	swapf	TMP_STATUS, W			; 
	movwf	STATUS					; restore STATUS
	swapf	TMP_WREG, F				; 
	swapf	TMP_WREG, W				; restore W
	PORT_cf					;
	retfie					; interrupt handled

;---------------------------------------------------------------------------
; UI primitives
;---------------------------------------------------------------------------

UI_INIT						; UI init routine
	movlw	B'00010000'
	movwf	UI_PORT
	bsf		STATUS,RP0
	movlw	UI_TRIS
	movwf	TRISA			; setup UI PORT
	movlw	AP_OPTION_REG	
 	movwf	OPTION_REG		; setup timer prescaler and timer mode
	bcf		STATUS,RP0
	clrf	TMR0
	TMR0_cf
	movlw	.32
	movwf	UI_TICK			; 32 * 256 * 256 blink on reset start (4 sec)
	return

UI_SETBLINKER
	movwf	UI_BLINKER
	btfss	UI_ERROR		; if LED is on (0), keep LEDS in current state
	return					; 
	iorlw	0x10			; otherwise keep LED err off
	movwf	UI_PORT			; drive LEDS
	return

UI_BLINK					; LED blinking routine
	TMR0_cf					; clear TMR0 IF
	decfsz	UI_TICK, F		; count ticks
	return					; not yet expired
	bcf		UI_ERROR		; error flag is reset
	bsf		UI_TICK, .1		; 2 * 256 * 256 blink period (~0.13 sec)
	rrf		UI_BLINKER, W	; ensure bits are not lost while rotating
	rrf		UI_BLINKER, F	; rotate 
	rrf		UI_BLINKER, W	; 
	rrf		UI_BLINKER, F	; two bits
	movf	UI_BLINKER, W	; get the result
	iorlw	B'00010000'		; set error bit (LED ERR off)
	movwf	UI_PORT			; indicate
	return

UI_SHOWERR
	bsf		UI_ERROR
	bcf		UI_LEDERR		; light up LED ERR
	bsf		UI_TICK, .4		; add additional ticks to show the error
	return

UI_SHOWSELECTION
	rlf		UI_SELECTION, W
	movwf	UI_BLINKER
	rlf		UI_BLINKER, W
	andlw	B'00001100'			; turn on LED ERR
	movwf	UI_PORT
	return

UI_getSEL macro				; read speed selection switch
	movf	UI_PORT, W
	andlw	0x03	
	movwf	UI_SELECTION
  endm
 
UI_showERR macro			
	call	UI_SHOWERR			
  endm
  
UI_clrERR macro
	bsf		UI_LEDERR		; turn OFF LED err (RB4 is open collector output)
	bcf		UI_ERROR
  endm
  
UI_setERR macro				; turn on LED ERR
	bcf		UI_LEDERR
	bsf		UI_ERROR
  endm
  
UI_indREADY macro
	movf	UI_SELECTION, W
	andlw	0x03	
	addlw	.12
	call	AP_GETBLINKER
	call	UI_SETBLINKER
  endm

UI_indWAIT macro
	movf	UI_SELECTION, W
	andlw	0x03	
	addlw	.8
	call	AP_GETBLINKER
	call	UI_SETBLINKER
  endm

UI_indRCV macro
	movf	UI_SELECTION, W
	andlw	0x03	
	addlw	.4
	call	AP_GETBLINKER
	call	UI_SETBLINKER
  endm

UI_indSEND macro
	movf	UI_SELECTION, W
	andlw	0x03	
	call	AP_GETBLINKER
	call	UI_SETBLINKER
  endm

UI_showSTART macro			; indicate RESET
	call	UI_SHOWSELECTION
  endm
	
;---------------------------------------------------------------------------
; UTRS application primitives

AP_GETFLAG						; converts UI_SLECTION to AP_FLAGS
	movf	UI_SELECTION, W
	andlw	0x03		
	addwf	PCL, F
	retlw	0x80
	retlw	0x40
	retlw	0x20
	retlw	0x10

AP_setFLAGS macro
	call	AP_GETFLAG
	iorwf	UI_SELECTION, F
  endm

AP_INIT							; initialization routine
	RS_setBUFFER
	clrf	RS_COUNTER
	movlw	RS_BUFFSIZE		
clrmem							; clear buffer	
	clrf	INDF
	incf	FSR, F
	addlw	-.1
	btfss	STATUS, Z
	goto	clrmem
	RS_setBUFFER
	
AP_SETDATA						; data preparation routine
	movlw	0x1F
	movwf	AP_DATA
	return

AP_LOADDATA
	incfsz	AP_DATA, F
	goto	_1	
	call	AP_SETDATA			
	incf	AP_DATA, F
_1	movf	AP_DATA, W
	return

AP_GETBLINKER
	addwf	PCL, F
	; SEND (DTR=on, RTS=off)
	retlw	B'11010101'		; 225K
	retlw	B'11011101'		; 112K
	retlw	B'11010111'		;  56K
	retlw	B'11011111'		;  37K
	; RECEIVE (DTR=off, RTS=on)
	retlw	B'10101011'		; 225K
	retlw	B'10111011'		; 112K
	retlw	B'10101111'		;  56K
	retlw	B'10111111'		;  37K
	; WAIT (DTR=on, RTS=on)
	retlw	B'10000100'		; 225K
	retlw	B'10011001'		; 112K
	retlw	B'10100101'		;  56K
	retlw	B'10010101'		;  37K
	; READY (DTR=off, RTS=off)
	retlw	B'00000110'		; 225K
	retlw	B'00000101'		; 112K
	retlw	B'00001010'		;  56K
	retlw	B'00001111'		;  37K

AP_SEND
	btfsc	AP_SPEED4
	goto	RS_SEND225K
	btfsc	AP_SPEED3
	goto	RS_SEND112K
	btfss	AP_SPEED2
	goto	s37
	call	RS_SEND56K			; 56 K needs some delay for stop bit
	goto	STOPBIT56K			; timing stop bit
s37	call	RS_SEND37K			; 37 K needs some delay for stop bit
	goto	STOPBIT37K			; timing stop bit
	
;---------------------------------------------------------------------------
; UTRS application code	
; UTRS performs two test cases:
; TEST_TX - trasmits sequence of bytes 20h-FFh
; TEST_RX - reads data into buffer and sends them back
; Tests are allowed if both RTS and DTR are SPACE
; TEST_RX is initiated by SPACE on RTS 
; TEST_TX is initiated by SPACE on DTR
;  
;------------------------------
; DTR\RTS |  MARK   |  SPACE  |
;---------+---------+---------|
;  MARK   |  RX     |   RX    |
;  SPACE  |  TX     |  WAIT   |
;------------------------------

DTR_skpSPACE macro
	btfss	AP_DTR
 endm

DTR_skpMARK macro
	btfsc	AP_DTR
 endm

TIMEOUT						; Timeout entry
	di
	call	RS_INIT
	call	UI_INIT
	UI_showERR				; indicate timeout
	call	RS_SENDBUFF		; send if something is in the buffer
	RS_setBUFFER
	clrf	RS_COUNTER		
	ei
	goto	WRTS			

START						; reset entry
	di
	call	RS_INIT
	call	UI_INIT
	call	AP_INIT
	UI_getSEL				; read speed selector
	AP_setFLAGS
	UI_showSTART			; indicate start (power-on, reset)		
	TMR0_ei
	ei

WAITREADY
	UI_indWAIT
	call	AP_SETDATA		; reset data byte to send		
WDTR ; waiting for DTR & ~RTS
	clrwdt
	DTR_skpSPACE
	goto	WAITRTS			; ~DTR found
	RTS_skpSPACE
	goto	DO_TX			; ~RTS found
	goto	WDTR			; DTR & RTS found
	
WAITRTS ; waiting for RTS & ~DTR
	UI_indREADY
WRTS ; waiting for RTS & ~DTR
	clrwdt
	RTS_skpSPACE
	goto	_RTS			; ~RTS & ~DTR found
;	 signal recieved, wait for incoming data
	UI_indRCV				; indicate RECIEVE
	; prepare for data reading
	movf	INTCON, W		; load INTCON register
	movwf	TMP_INTCON		; save INTCON
	movlw	RS_INTCON
	andwf	INTCON,F		; disable all interrupts, clear INTF
rxloop
	call	RS_READBUFF		; read data into the buffer
	RS_setBUFFER			; setup data buffer
	call	RS_SENDBUFF		; send data from the buffer
	RS_setBUFFER			; setup data buffer
	call	UI_BLINK		; blink to indicate we are alive
	RTS_skpMARK				; is RTS kept space?		
	goto	rxloop			; yes, read again
done	; no, RX done
	movf	TMP_INTCON, W	; load previous value
	andlw	~RS_INTCON		; mask interruption flags
	iorwf	INTCON, F		; enable what was enabled before	
	INT_di					; disable RX interrupt
	goto	WAITRTS			; wait for RTS again

	call	UI_BLINK
	goto	WRTS
_RTS
	DTR_skpMARK
	goto	WAITREADY		; DTR found
	goto	WRTS			; ~RTS & ~DTR found	


DO_TX		; Send test data
	UI_indSEND
tx	clrwdt	
	call	AP_LOADDATA		; load test data into W
	di
	call	AP_SEND			; send with requested speed
	ei	
	RTS_skpMARK				; test if RTS is kept MARK
	goto	WAITREADY		; request for a new test detected
	DTR_skpMARK				; is SPACE on DTR yet?
	goto	tx				; continue sending data
	goto	WDTR			; ~DTR & ~RTS condition detected, suspend sending
	
STOPBIT37K	
	RS_delay .8					; Delay for 29 IC
	RS_delay .7
STOPBIT56K						; Delay for 14 instruction cycles
	RS_delay .5					
	RS_delay .5					
	return	

RS  CODE
RS_SEND
RS_SEND37K
	RS_bodySEND .38400, RS_BITCOUNT, TX, 0
	return
RS_SEND56K
	RS_bodySEND .57600, RS_BITCOUNT, TX, 0
	return	
RS_SEND112K
	RS_bodySEND .115200, RS_BITCOUNT, TX, 0
	return	
RS_SEND225K
	RS_bodySEND .230400, RS_BITCOUNT, TX, 0
	return	
;RS_SEND450K
;	RS_bodySEND .460800, RS_BITCOUNT, TX, 0
;	return
	
;---------------------------------------------------------------------------
RS_READBUFF					; READ incoming data to buffer
	INT_cf					; clear INT FLAG	
	INT_ei					; enable INT interrupt
	ei						; enable interrupts
	CTS_setSPACE			; transmission shall begin
	TMR0_clr				; clear timer
	TMR0_cf					; clear timer expiration flag
rx	clrwdt
	RTS_skpSPACE			; is RTS kept SPACE
	goto	suspend				; no, suspend reception
	tstf	RS_COUNTER		; test is something received
	skpnz					
	goto	rx				; if no bytes yet received, just wait
	TMR0_skpc				; if timer expired
	goto	suspend			; suspend reception
	movlw	RS_MAXSKID		; watch buffer overflow
	subwf	RS_COUNTER, W	; RS_COUNTER - RS_MAXSKID
	skpc					; if( RS_COUNTER - RS_MAXSKID <= 0)
	goto	rx				; continue reception
suspend	
	CTS_setMARK				; request suspending reception
	TMR0_clr				; clear timer
	; relaxation period 2*bit length after last byte recived
xr	clrwdt
	movlw	RS_BUFFSIZE - 1
	subwf	RS_COUNTER, W	; RS_COUNTER - (RS_BUFFSIZE - 1)
	skpnc					; if( RS_COUNTER - (RS_BUFFSIZE - 1) > 0)
	goto	readdone					
	RTS_skpSPACE			; wait if space is kept SPACE
	goto	readdone					
	btfss	TMR0, 2			; time out = 4 timer ticks (1024 IC) ~ 1 ms, good for fast PC's
	;TMR0_skps				; timeout = 256	 timer ticks ~ 0.25 s
	goto	xr
readdone
	di
	INT_di					; enable INT interrupt
	return		
;---------------------------------------------------------------------------
RS_SENDBUFF			; sending back recieved data
	tstf	RS_COUNTER
	skpnz
	return
sendbyte
	movf	INDF, W			; send a byte from the buffer
	di
	call	AP_SEND			; at application-defined speed
	ei	
	incf	FSR, F
	decfsz	RS_COUNTER, F
	goto    sendbyte
	return

;---------------------------------------------------------------------------

INT_SPEED3							; 115200
	swapf	STATUS, W				; 16: load STATUS
	movwf	TMP_STATUS				; 17: save STATUS
	btfss	AP_SPEED3				; 19: test if is the selected speed is 112K
	goto	INT_SPEED2				; 20: no, it is not
	clrc							; 21: prepare C for the next shift
	rrf		INDF, F					; 22: bit read at IC 12 is actually the first data bit
	RS_bodyRECEIVE .115200, RS_BITCOUNT, RX, 0, INDF, .21, -.3

INT_RSDONE
	TMR0_clr						; 75: clear timer
	incf	FSR, F					; 76: forward buffer pointer
	incf	RS_COUNTER, F			; 77: count this byte
	swapf	TMP_STATUS, W			; 78
	movwf	STATUS					; 79: restore STATUS
	swapf	TMP_WREG, F				; 80:
	swapf	TMP_WREG, W				; 81: restore W
	INT_cf							; 82: interrupt handled
	retfie							; 83: leave routine - 2 cycles of one stop bit left

INT_SPEED2							; 57600
	btfss	AP_SPEED2				; 22: test if is the selected speed is 112K
	goto	INT_SPEED1				; 23: no, it is not
	clrf	INDF					; 24: discard whatever was recived
	RS_bodyRECEIVE .57600, RS_BITCOUNT, RX, 0, INDF, .24, -.6
	goto	INT_RSDONE

INT_SPEED1							; 38400
	clrf	INDF					; 25: discard whatever was recived
	RS_bodyRECEIVE .38400, RS_BITCOUNT, RX, 0, INDF, .25, -.6
	goto	INT_RSDONE

	
;---------------------------------------------------------------------------
	RS_delays
;---------------------------------------------------------------------------
  
RS_INIT
	INT_di				; disable port interrupts
	movlw	0xFF		; drive high all pins on port B
	movwf	PORTB
	movlw	RS_TRIS
	bsf		STATUS,RP0
	movwf	TRISB
	bcf		STATUS,RP0
;	movlw	0xFF
;	movwf	PORTB
    retlw	0
	END
